/* bmp32.cpp
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <malloc.h>

namespace dib
{
	int GetBytesPerPixel(int depth)
	{
		return (depth==32 ? 4 : 3);
	}

	int GetBytesPerRow(int width, int depth)
	{
		int bytesPerPixel = GetBytesPerPixel(depth);
		int bytesPerRow = ((width * bytesPerPixel + 3) & ~3);

		return bytesPerRow;
	}

	int GetBitmapBytes(int width, int height, int depth)
	{
		return height * GetBytesPerRow(width, depth);
	}

	int GetBitmapBytes(const BITMAPINFOHEADER *bmih)
	{
		return GetBitmapBytes(bmih->biWidth, bmih->biHeight, bmih->biBitCount);
	}

	RGBTRIPLE *GetPixelPtr24(BYTE *pixelData, int width, int height, int depth, int x, int y)
	{
		if (depth != 24)
			return NULL;

		if (x < 0 || x >= width || y < 0 || y >= height)
			return NULL;

		int bytesPerRow = GetBytesPerRow(width, depth);
		int bytesPerPixel = GetBytesPerPixel(depth);

		int rowOffset = y * bytesPerRow;	// Gets us to the correct row.
		int colOffset = x * bytesPerPixel;	// Gets us to the correct column.

		int pixelOffset = rowOffset + colOffset;

		return (RGBTRIPLE *)(pixelData + pixelOffset);
	}

	const RGBQUAD *GetPixelPtr32(const BYTE *pixelData, int width, int height, int depth, int x, int y)
	{
		if (depth != 32)
			return NULL;

		if (x < 0 || x >= width || y < 0 || y >= height)
			return NULL;

		int bytesPerRow = GetBytesPerRow(width, depth);
		int bytesPerPixel = GetBytesPerPixel(depth);

		int rowOffset = y * bytesPerRow;	// Gets us to the correct row.
		int colOffset = x * bytesPerPixel;	// Gets us to the correct column.

		int pixelOffset = rowOffset + colOffset;

		return (RGBQUAD *)(pixelData + pixelOffset);
	}
};	// namespace dib

void ExtractRGB(const BITMAPINFOHEADER *bmihSource, const char *rgb_file)
{
	printf("Extracting RGB channels...\n");

	const BYTE *sourcePixels = ((const BYTE *)bmihSource) + bmihSource->biSize;
	int imageBytes = dib::GetBitmapBytes(bmihSource->biWidth, bmihSource->biHeight, 24);

	// Write that out to a .BMP file.  A .BMP file starts with a BITMAPFILEHEADER
	BITMAPFILEHEADER bmfh;
	bmfh.bfType = MAKEWORD('B','M');
	bmfh.bfSize = imageBytes + sizeof(BITMAPFILEHEADER) + 0x28;
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;
	bmfh.bfOffBits = 0x0036;

	// That's then followed by a BITMAPINFO structure.
	BITMAPINFO bmi;
	bmi.bmiHeader.biSize = 0x28;
	bmi.bmiHeader.biWidth = bmihSource->biWidth;
	bmi.bmiHeader.biHeight = bmihSource->biHeight;
	bmi.bmiHeader.biPlanes = 1;
	bmi.bmiHeader.biBitCount = 24;
	bmi.bmiHeader.biCompression = BI_RGB;
	bmi.bmiHeader.biSizeImage = imageBytes;
	bmi.bmiHeader.biXPelsPerMeter = 0x2e23;
	bmi.bmiHeader.biYPelsPerMeter = 0x2e23;
	bmi.bmiHeader.biClrUsed = 0;
	bmi.bmiHeader.biClrImportant = 0;

	// Then it's followed by the pixel data.
	BYTE *destPixels = (BYTE *)malloc(imageBytes);
	memset(destPixels, 0, imageBytes);

	// Then there are num_rows * num_cols B records.  Generally num_rows == 1, num_cols == 1201.
	for (int y = 0; y < bmihSource->biHeight; ++y)
	{
		for (int x = 0; x < bmihSource->biWidth; ++x)
		{
			const RGBQUAD *source = dib::GetPixelPtr32(sourcePixels, bmihSource->biWidth, bmihSource->biHeight, 32, x, y);
			RGBTRIPLE *dest = dib::GetPixelPtr24(destPixels, bmihSource->biWidth, bmihSource->biHeight, 24, x, y);

			dest->rgbtRed = source->rgbRed;
			dest->rgbtGreen = source->rgbGreen;
			dest->rgbtBlue = source->rgbBlue;
		}
	}

	HANDLE hBmp = CreateFile(rgb_file, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
	if (hBmp != INVALID_HANDLE_VALUE)
	{
		DWORD bytesWritten;
		WriteFile(hBmp, &bmfh, sizeof(BITMAPFILEHEADER), &bytesWritten, NULL);
		WriteFile(hBmp, &bmi, sizeof(BITMAPINFOHEADER), &bytesWritten, NULL);
		WriteFile(hBmp, destPixels, imageBytes, &bytesWritten, NULL);
		CloseHandle(hBmp);
	}
	else
		fprintf(stderr, "CreateFile failed, '%s', error = %d.\n", rgb_file, GetLastError());
	free(destPixels);
}

void ExtractAlpha(const BITMAPINFOHEADER *bmihSource, const char *alpha_file)
{
	printf("Extracting Alpha channels...\n");

	const BYTE *sourcePixels = ((const BYTE *)bmihSource) + bmihSource->biSize;
	int imageBytes = dib::GetBitmapBytes(bmihSource->biWidth, bmihSource->biHeight, 24);

	// Write that out to a .BMP file.  A .BMP file starts with a BITMAPFILEHEADER
	BITMAPFILEHEADER bmfh;
	bmfh.bfType = MAKEWORD('B','M');
	bmfh.bfSize = imageBytes + sizeof(BITMAPFILEHEADER) + 0x28;
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;
	bmfh.bfOffBits = 0x0036;

	// That's then followed by a BITMAPINFO structure.
	BITMAPINFO bmi;
	bmi.bmiHeader.biSize = 0x28;
	bmi.bmiHeader.biWidth = bmihSource->biWidth;
	bmi.bmiHeader.biHeight = bmihSource->biHeight;
	bmi.bmiHeader.biPlanes = 1;
	bmi.bmiHeader.biBitCount = 24;
	bmi.bmiHeader.biCompression = BI_RGB;
	bmi.bmiHeader.biSizeImage = imageBytes;
	bmi.bmiHeader.biXPelsPerMeter = 0x2e23;
	bmi.bmiHeader.biYPelsPerMeter = 0x2e23;
	bmi.bmiHeader.biClrUsed = 0;
	bmi.bmiHeader.biClrImportant = 0;

	// Then it's followed by the pixel data.
	BYTE *destPixels = (BYTE *)malloc(imageBytes);
	memset(destPixels, 0, imageBytes);

	// Then there are num_rows * num_cols B records.  Generally num_rows == 1, num_cols == 1201.
	for (int y = 0; y < bmihSource->biHeight; ++y)
	{
		for (int x = 0; x < bmihSource->biWidth; ++x)
		{
			const RGBQUAD *source = dib::GetPixelPtr32(sourcePixels, bmihSource->biWidth, bmihSource->biHeight, 32, x, y);
			RGBTRIPLE *dest = dib::GetPixelPtr24(destPixels, bmihSource->biWidth, bmihSource->biHeight, 24, x, y);

			dest->rgbtRed = source->rgbReserved;
			dest->rgbtGreen = source->rgbReserved;
			dest->rgbtBlue = source->rgbReserved;
		}
	}

	HANDLE hBmp = CreateFile(alpha_file, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
	if (hBmp != INVALID_HANDLE_VALUE)
	{
		DWORD bytesWritten;
		WriteFile(hBmp, &bmfh, sizeof(BITMAPFILEHEADER), &bytesWritten, NULL);
		WriteFile(hBmp, &bmi, sizeof(BITMAPINFOHEADER), &bytesWritten, NULL);
		WriteFile(hBmp, destPixels, imageBytes, &bytesWritten, NULL);
		CloseHandle(hBmp);
	}
	else
		fprintf(stderr, "CreateFile failed, '%s', error = %d.\n", alpha_file, GetLastError());
	free(destPixels);
}

int main(int argc, char *argv[])
{
	if (argc < 4)
	{
		fprintf(stderr, "Usage:\nbmp32 32bpp.bmp 24bpp.bmp alpha.bmp\n");
		return 1;
	}

	// Load the file specified and separate the RGB channels into one file and the A channel into a different one.
	HANDLE hFile = CreateFile(argv[1], GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (hFile != INVALID_HANDLE_VALUE)
	{
		// Allocate a chunk of memory and load the file into it.
		ULARGE_INTEGER uli;
		uli.LowPart = GetFileSize(hFile, &uli.HighPart);

		// We can't actually load that much data:
		BYTE *pSource = new BYTE[uli.LowPart];

		DWORD bytesRead;
		BOOL ok = ReadFile(hFile, pSource, uli.LowPart, &bytesRead, NULL);
		if (ok)
		{
			const BITMAPFILEHEADER *bmfhSource = (const BITMAPFILEHEADER *)pSource;
			const BITMAPINFO *bmiSource = (const BITMAPINFO *)(pSource + sizeof(BITMAPFILEHEADER));
			const BITMAPINFOHEADER *bmihSource = (const BITMAPINFOHEADER *)(&bmiSource->bmiHeader);

			// If the bitmap is 32-bpp, then we need to grab the RGB channel and dump it into a different file.
			printf("%s is a %d-bpp bitmap.\n", argv[1], bmihSource->biBitCount);
			if (bmihSource->biBitCount == 32)
			{
				ExtractRGB(bmihSource, argv[2]);
				ExtractAlpha(bmihSource, argv[3]);
			}
		}

		delete [] pSource;
		CloseHandle(hFile);
	}
	else
		fprintf(stderr, "Failed to open source file '%s'\n", argv[1]);

	return 0;
}
